#-----------------------------------------------------------------------
# Aplicação de bootstrap. Estudo de caso com modelos não lineares.

library(alr3)
str(turk0)

plot(Gain ~ A, data = turk0)

# Lista com os valores iniciais.
start <- list(Int = 625, Ass = 180, Mei = 0.1)

# Adiciona as curvas.
with(start, {
    curve(Int + Ass * A/(Mei + A),
          xname = "A", add = TRUE, col = 2)
    curve(Int + Ass * (1 - 2^(-A/Mei)),
          xname = "A", add = TRUE, col = 4)})

# Ajuste do primeiro modelo.
n0 <- nls(Gain ~ Int + Ass * A/(Mei + A),
          data = turk0,
          start = start)

# Ajuste do segundo modelo.
n1 <- nls(Gain ~ Int + Ass * (1 - 2^(-A/Mei)),
          data = turk0,
          start = start)

cbind(coef(n0), coef(n1))

#-----------------------------------------------------------------------
# Criando funções que fazem reamostragem dos dados e ajustam o modelo.

n <- nrow(turk0)
i <- 1:n

myboot0 <- function() {
    n0 <- nls(Gain ~ Int + Ass * A/(Mei + A),
              data = turk0[sample(i, n, replace = TRUE), ],
              start = start)
    coef(n0)
}

myboot1<- function() {
    n0 <- nls(Gain ~ Int + Ass * (1 - 2^(-A/Mei)),
              data = turk0[sample(i, n, replace = TRUE), ],
              start = start)
    coef(n0)
}

#-----------------------------------------------------------------------
# Replicando.

b0 <- replicate(999, myboot0())
b1 <- replicate(999, myboot1())

b0 <- cbind(b0, coef(n0))
b1 <- cbind(b1, coef(n1))
str(b0)

# Vício.
cbind(n0 = rowMeans(b0) - coef(n0),
      n1 = rowMeans(b1) - coef(n1))

# Variância.
round(cbind(n0 = apply(b0, MARGIN = 1, var),
            n1 = apply(b1, MARGIN = 1, var)), digits = 5)

# Função para calcular o erro quadrático médio.
eqm <- function(x) {
    sum((x - tail(x, 1))^2)/length(x)
}

# Erro quadrático médio.
round(cbind(n0 = apply(b0, 1, FUN = eqm),
            n1 = apply(b1, 1, FUN = eqm)), digits = 5)

#-----------------------------------------------------------------------
# Curvas ajustadas de onde é possível determinar uma banda de confiança.

formula(n0)
formula(n1)

pairs(t(b0))
pairs(t(b1))

par(mfrow = c(1, 2))
plot(Gain ~ A, data = turk0)
apply(b0, MARGIN = 2,
      FUN = function(b) {
          curve(b[1] + b[2] * x/(b[3] + x),
                add = TRUE, col = rgb(0, 0, 1, 0.25), n = 51)
          invisible()
      })
plot(Gain ~ A, data = turk0)
apply(b1, MARGIN = 2,
      FUN = function(b) {
          curve(b[1] + b[2] * (1 - 2^(-x/b[3])),
                add = TRUE, col = rgb(0, 0, 1, 0.25), n = 51)
          invisible()
      })

#-----------------------------------------------------------------------
# Estimativa do valor da função em x = 0.2.

p0 <- apply(b0, MARGIN = 2,
            FUN = function(b){
                b[1] + b[2] * 0.2/(b[3] + 0.2)
            })
p1 <- apply(b1, MARGIN = 2,
            FUN = function(b){
                b[1] + b[2] * (1 - 2^(-0.2/b[3]))
            })

# Desvio padrão de \hat{y} em x = 0.2.
cbind(sd(p0), sd(p1))

# Estimativa de "vida" 90%.
q <- 0.9
m0 <- b0[3, ] * q/(1 - q)
m1 <- -log2(1 - q) * b1[3, ]

cbind(n0 = summary(m0), n1 = summary(m1))

# Distribuição do x_{90%}.
par(mfrow = c(1, 2))
plot(density(m0))
rug(m0)
abline(v = coef(n0)["Mei"] * q/(1 - q))
plot(density(m1))
rug(m1)
abline(v = -log2(1 - q) * coef(n1)["Mei"])
layout(1)

plot(Gain ~ A, turk0, xlim = c(0, 2), ylim = c(590, 850))
with(as.list(coef(n0)), {
    curve(Int + Ass * A/(Mei + A),
          xname = "A", add = TRUE, col = 2)
    y90 <- Int + q * Ass
    x90 <- Mei * q/(1 - q)
    abline(h = y90, v = x90, col = 2, lty = 2)
    points(x90, y90, pch = 19, col = 2)
    rug(m0, col = 2)
})
with(as.list(coef(n1)), {
    curve(Int + Ass * (1 - 2^(-A/Mei)),
          xname = "A", add = TRUE, col = 4)
    y90 <- Int + q * Ass
    x90 <- -log2(1 - q) * Mei
    abline(h = y90, v = x90, col = 4, lty = 2)
    points(x90, y90, pch = 19, col = 4)
    rug(m1, col = 4)
})

#-----------------------------------------------------------------------
# Usando o pacote boot.

library(boot)
library(latticeExtra)

help(boot, help_type = "html")

# dput(as.list(coef(n1)))
start <- list(Int = 622.958054146589,
              Ass = 178.251908347242,
              Mei = 0.097321927254495)

# Criar uma função com dois argumentos: o data.frame original e um vetor
# que vai representar o índice das linhas usado para reamostrar dos
# dados.
fitmodel <- function(dataset, index) {
    n0 <- nls(Gain ~ Int + Ass * (1 - 2^(-A/Mei)),
              data = dataset[index, ],
              start = start)
    c(coef(n0), s2 = deviance(n0)/df.residual(n0))
}

set.seed(321)
b0 <- boot(data = turk0,
           statistic = fitmodel,
           R = 1999)

class(b0)
methods(class = class(b0))

summary(b0)

# Como são obtidos.
theta <- b0$t

# Estimativas como conjunto de dados original.
b0$t0

# Vício.
apply(theta, MARGIN = 2, FUN = mean) - b0$t0

# Desvio-padrão boostrap.
apply(theta, MARGIN = 2, FUN = sd)

# Mediana bootstrap.
apply(theta, MARGIN = 2, FUN = median)

st <- summary(b0)
str(st)

# Gráfico de pares.
pairs(b0)
pairs(b0$t)

vcov(b0)
cov2cor(vcov(b0))

# Extrair os vetores de estativativas bootstrap.
str(b0)
colnames(b0$t) <- names(b0$t0)
best <- stack(as.data.frame(b0$t))
names(best)

st

densityplot(~values | ind,
            data = best,
            scales = "free",
            as.table = TRUE) +
    layer(panel.abline(v = st$original[which.packet()] +
                           c(0, st$bootBias[which.packet()]),
                       col = c(1, 2),
                       lty = 2))

# Intervalos de confiança.
confint(b0, type = "norm")
confint(b0, type = "basic")
confint(b0, type = "perc")
confint(b0, type = "bca")

# Todos métodos (apenas para o primeiro parâmetro).
bci <- boot.ci(b0, type = "all", index = c(2, 2))
bci <- boot.ci(b0, type = "all", index = c(3, 3))
bci <- boot.ci(b0, type = "all", index = c(4, 4))
bci

str(bci)

#-----------------------------------------------------------------------
# Abrir as funções para estudar o código fonte.

# Abrir o código fonte para entender.
boot.ci

# Abrir cada um dos métodos de calculo de IC.
norm.ci
boot:::basic.ci
boot:::stud.ci
boot:::perc.ci
boot:::bca.ci

#-----------------------------------------------------------------------
# Fazendo o debugging das funções.

# debug(norm.ci)
# norm.ci(b0)
# undebug(norm.ci)
#
# debug(confint)
# confint(b0, type = "basic")
# undebug(confint)

#-----------------------------------------------------------------------
#-----------------------------------------------------------------------
# Estudo de simulação sobre o desempenho dos tipos de intervalo.

#--------------------------------------------
# 1. Gerar observações assumindo um modelo linear com resposta Poisson.

simul_model <- function(n = 50,
                        x = rep(seq(0, 3, length.out = 10), times = 5),
                        beta = c(-0.5, 1)) {
    eta <- beta[1] + beta[2] * x
    mu <- exp(eta)
    y <- rpois(n = n, lambda = mu)
    return(data.frame(x = x, y = y))
}

# Vendo o resultado do modelo.
with(simul_model(),
     expr = {
         plot(y ~ x)
     })

plot(y ~ x, data = simul_model())

#--------------------------------------------
# 2. Para cada novo conjunto de dados, obter os IC com cada método
# usando um R = 999.

# Função que ajusta e retorna as estatísticas de interesse.
fitmodel <- function(dataset, index) {
    g0 <- glm(y ~ x,
              data = dataset[index, ],
              family = poisson(link = log))
    return(c(coef(g0), deviance(g0)))
}

# Valores usados para os parâmetros.
beta <- c(-0.5, 1)

# Corre as simulações bootstrap (reamostragens).
btst <- boot(data = simul_model(beta = beta),
             statistic = fitmodel,
             R = 999)

# Obtém os intervalos de confiança para uma das estatísticas.
bci <- boot.ci(btst,
               type = "all",
               index = c(2, 2))
bci
str(bci)

# Extração dos limites do IC.
ics <- sapply(bci[-(1:3)],
              FUN = function(x) {
                  # Pega as duas últimas colunas.
                  x[, ncol(x) - 1:0]
              })
ics

# Verificar se contém o não o valor do parâmetro.
inside <- apply(ics,
                MARGIN = 2,
                FUN = function(lim) {
                    prod(lim - beta[2]) < 0
                })
inside

# Comprimento de cada IC.
apply(ics,
      MARGIN = 2,
      FUN = diff)

#--------------------------------------------
# 3. Repetir isso 100 vezes.

proc <- function(..., R = 999) {
    btst <- boot(data = simul_model(...),
                 statistic = fitmodel,
                 R = R)
    bci <- boot.ci(btst,
                   type = "all",
                   index = c(2, 2))
    ics <- sapply(bci[-(1:3)],
                  FUN = function(x) {
                      x[, ncol(x) - 1:0]
                  })
    inside <- apply(ics,
                    MARGIN = 2,
                    FUN = function(lim) {
                        prod(lim - beta[2]) < 0
                    })
    return(inside)
}

system.time(proc(beta = c(-0.5, 1),
                 x = rep(seq(0, 3, by = 0.5), 3),
                 n = 21))

N <- 5
result <- replicate(N,
                    proc(beta = c(-0.5, 1),
                         x = rep(seq(0, 3, by = 0.5), 3),
                         n = 21))

str(result)
# result

#--------------------------------------------
# 4. Verificar a taxa de cobertura dos intervalos de confiança.

# apply()
rowSums(result)/N

# Salva imagem com os objetos criados.
save.image(file = "my_results.RData")

#-----------------------------------------------------------------------

PaulaTb3.12 <-
    structure(list(
        vol = c(3.7, 3.5, 1.25, 0.75, 0.8, 0.7, 0.6, 1.1, 0.9, 0.9, 0.8,
                0.55, 0.6, 1.4, 0.75, 2.3, 3.2, 0.85, 1.7, 1.8, 0.4,
                0.95, 1.35, 1.5, 1.6, 0.6, 1.8, 0.95, 1.9, 1.6, 2.7,
                2.35, 1.1, 1.1, 1.2, 0.8, 0.95, 0.75, 1.3),
        razao = c(0.825, 1.09, 2.5, 1.5, 3.2, 3.5, 0.75, 1.7, 0.75,
                  0.45, 0.57, 2.75, 3, 2.33, 3.75, 1.64, 1.6, 1.415,
                  1.06, 1.8, 2, 1.36, 1.35, 1.36, 1.78, 1.5, 1.5, 1.9,
                  0.95, 0.4, 0.75, 0.03, 1.83, 2.2, 2, 3.33, 1.9, 1.9,
                  1.625),
        resp = c(1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
                 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 1,
                 0, 0, 1)),
        .Names = c("vol", "razao", "resp"),
        row.names = c(NA, 39L),
        class = "data.frame")

layout(1)

#  Visualização dos dados.
xyplot(resp ~ vol,
       data = PaulaTb3.12,
       groups = resp)

#  Ajuste do modelo considerando resposta binária.
m0 <- glm(resp ~ vol,
          data = PaulaTb3.12,
          family = binomial)
summary(m0)

# Estimativa da DL_50 a partir do modelo ajustado.
dl <- -coef(m0)[1]/coef(m0)[2]

# Curva ajustada.
plot(resp ~ vol, data = PaulaTb3.12)
curve(m0$family$linkinv(coef(m0)[1] + coef(m0)[2] * x),
      add = TRUE)
abline(v = dl, h = 0.5, lty = 2)

dl_50 <- function(dataset, index) {
    m0 <- glm(resp ~ vol,
              data = dataset[index, ],
              family = binomial)
    dl <- -coef(m0)[1]/coef(m0)[2]
    return(dl)
}

dl50 <- boot(data = PaulaTb3.12,
             statistic = dl_50,
             R = 2999)

summary(dl50)

hist(dl50)

boot.ci(dl50,
        type = "all",
        index = c(1, 1))

#-----------------------------------------------------------------------
